#ifndef _PACKET_PRIV_H
#define _PACKET_PRIV_H

#ifdef __cplusplus
extern "C"
{
#endif

#include <stdint.h>
#include <stdio.h>

#include "log.h"
#include "stellar/tuple.h"
#include "stellar/packet.h"

#define PACKET_MAX_LAYERS 32
#define PACKET_LOG_ERROR(format, ...) LOG_ERROR("packet", format, ##__VA_ARGS__)
#define PACKET_LOG_WARN(format, ...) LOG_WARN("packet", format, ##__VA_ARGS__)
#define PACKET_LOG_DEBUG(format, ...) void(0)
// #define PACKET_LOG_DEBUG(format, ...) LOG_DEBUG("packet", format, ##__VA_ARGS__)

enum ldbc_method
{
    LDBC_METHOD_HASH_INT_IP = 1,
    LDBC_METHOD_HASH_EXT_IP = 2,
    LDBC_METHOD_HASH_INT_IP_AND_EXT_IP = 3,
    LDBC_METHOD_HASH_INNERMOST_INT_IP = 4,
    LDBC_METHOD_HASH_INNERMOST_EXT_IP = 5,
};

enum packet_origin
{
    PACKET_ORIGIN_MARSIO = 0x1,    // packet data in mbuff      (eg: packet I/O mrzcpd mode)
    PACKET_ORIGIN_DUMPFILE = 0x2,  // packet data in pcap       (eg: packet I/O dumpfile mode)
    PACKET_ORIGIN_USERSTACK = 0x3, // packet data in user stack (eg: inject packet)
    PACKET_ORIGIN_USERHEAP = 0x4,  // packet data in user heap  (eg: ip reassembly)
};

struct packet
{
    struct packet_layer layers[PACKET_MAX_LAYERS];
    struct packet_layer *frag_layer; // fragment layer
    int8_t layers_used;
    int8_t layers_size;

    const char *data_ptr;
    uint16_t data_len;
    uint16_t trim_len; // trim eth padding

    void *origin_ctx; // mbuff or pcap pointer
    enum packet_action action;
    enum packet_origin origin;
};

// return innermost payload
const char *packet_parse(struct packet *pkt, const char *data, uint16_t len);
void packet_print_str(const struct packet *pkt);

// direction 1: E2I
// direction 0: I2E
uint64_t packet_get_hash(const struct packet *pkt, enum ldbc_method method, int direction);

/******************************************************************************
 * Utils
 ******************************************************************************/

#define MAX_ROUTE_CTX 64
struct route_ctx
{
    char data[MAX_ROUTE_CTX];
    int used;
};
void packet_set_route_ctx(struct packet *pkt, const struct route_ctx *ctx);
void packet_get_route_ctx(const struct packet *pkt, struct route_ctx *ctx);

void packet_set_origin(struct packet *pkt, enum packet_origin origin);
enum packet_origin packet_get_origin(const struct packet *pkt);

void packet_set_origin_ctx(struct packet *pkt, void *ctx);
void *packet_get_origin_ctx(const struct packet *pkt);

void packet_set_domain(struct packet *pkt, uint64_t domain);
uint64_t packet_get_domain(const struct packet *pkt);

void packet_set_ctrl(struct packet *pkt);
int packet_is_ctrl(const struct packet *pkt);

void packet_set_sid_list(struct packet *pkt, const struct sid_list *list);
void packet_get_sid_list(const struct packet *pkt, struct sid_list *list);
void packet_append_sid_list(struct packet *pkt, const struct sid_list *list);

struct packet *packet_new(uint16_t pkt_len);
void packet_free(struct packet *pkt);
struct packet *packet_dup(const struct packet *pkt);

void packet_set_session_id(struct packet *pkt, uint64_t sess_id);
void packet_set_direction(struct packet *pkt, enum packet_direction dir);

int packet_is_fragment(const struct packet *pkt);
int packet_get_fingerprint(const struct packet *pkt, char *buff, int size);

#ifdef __cplusplus
}
#endif

#endif
